@@ -14,7 +14,7 @@
 #![allow(clippy::cognitive_complexity)]
 #![allow(clippy::upper_case_acronyms)]
 #![cfg_attr(
-    not(any(feature = "libdeflater", feature = "zopfli")),
+    not(feature = "zopfli"),
     allow(irrefutable_let_patterns),
     allow(unreachable_patterns)
 )]
@@ -26,12 +26,11 @@ mod rayon;
 
 use crate::atomicmin::AtomicMin;
 use crate::colors::BitDepth;
-use crate::deflate::inflate;
+use crate::deflate::{crc32, inflate};
 use crate::evaluate::Evaluator;
 use crate::png::PngData;
 use crate::png::PngImage;
 use crate::reduction::*;
-use crc::{Crc, CRC_32_ISO_HDLC};
 use image::{DynamicImage, GenericImageView, ImageFormat, Pixel};
 use log::{debug, error, info, warn};
 use rayon::prelude::*;
@@ -46,8 +45,9 @@ use std::time::{Duration, Instant};
 pub use crate::colors::AlphaOptim;
 pub use crate::deflate::Deflaters;
 pub use crate::error::PngError;
+pub use crate::filters::RowFilter;
 pub use crate::headers::Headers;
-pub use indexmap::{IndexMap, IndexSet};
+pub use indexmap::{indexset, IndexMap, IndexSet};
 
 mod atomicmin;
 mod colors;
@@ -147,10 +147,10 @@ pub struct Options {
     ///
     /// Default: `false`
     pub preserve_attrs: bool,
-    /// Which filters to try on the file (0-5)
+    /// Which RowFilters to try on the file
     ///
-    /// Default: `0,5`
-    pub filter: IndexSet<u8>,
+    /// Default: `None,Sub,Entropy,Bigrams`
+    pub filter: IndexSet<RowFilter>,
     /// Whether to change the interlacing type of the file.
     ///
     /// `None` will not change the current interlacing type.
@@ -190,14 +190,12 @@ pub struct Options {
     pub strip: Headers,
     /// Which DEFLATE algorithm to use
     ///
-    /// Default: `Zlib`
+    /// Default: `Libdeflater`
     pub deflate: Deflaters,
-    /// Whether to use heuristics to pick the best filter and compression
+    /// Whether to use fast evaluation to pick the best filter
     ///
-    /// Intended for use with `-o 1` from the CLI interface
-    ///
-    /// Default: `false`
-    pub use_heuristics: bool,
+    /// Default: `true`
+    pub fast_evaluation: bool,
 
     /// Maximum amount of time to spend on optimizations.
     /// Further potential optimizations are skipped if the timeout is exceeded.
@@ -212,10 +210,7 @@ impl Options {
             1 => opts.apply_preset_1(),
             2 => opts.apply_preset_2(),
             3 => opts.apply_preset_3(),
-            4 => {
-                warn!("Level 4 is deprecated and is identical to level 3");
-                opts.apply_preset_4()
-            }
+            4 => opts.apply_preset_4(),
             5 => opts.apply_preset_5(),
             6 => opts.apply_preset_6(),
             _ => {
@@ -232,20 +227,18 @@ impl Options {
     // The following methods make assumptions that they are operating
     // on an `Options` struct generated by the `default` method.
     fn apply_preset_0(mut self) -> Self {
-        self.idat_recoding = false;
-        if let Deflaters::Zlib { compression, .. } = &mut self.deflate {
-            compression.clear();
-            compression.insert(3);
+        self.filter.clear();
+        if let Deflaters::Libdeflater { compression } = &mut self.deflate {
+            *compression = 5;
         }
         self
     }
 
     fn apply_preset_1(mut self) -> Self {
         self.filter.clear();
-        if let Deflaters::Zlib { strategies, .. } = &mut self.deflate {
-            strategies.clear();
+        if let Deflaters::Libdeflater { compression } = &mut self.deflate {
+            *compression = 10;
         }
-        self.use_heuristics = true;
         self
     }
 
@@ -254,53 +247,45 @@ impl Options {
     }
 
     fn apply_preset_3(mut self) -> Self {
-        for i in 1..5 {
-            self.filter.insert(i);
-        }
+        self.fast_evaluation = false;
+        self.filter = indexset! {
+            RowFilter::None,
+            RowFilter::Bigrams,
+            RowFilter::BigEnt,
+            RowFilter::Brute
+        };
         self
     }
 
-    fn apply_preset_4(self) -> Self {
+    fn apply_preset_4(mut self) -> Self {
+        if let Deflaters::Libdeflater { compression } = &mut self.deflate {
+            *compression = 12;
+        }
         self.apply_preset_3()
     }
 
     fn apply_preset_5(mut self) -> Self {
-        if let Deflaters::Zlib { compression, .. } = &mut self.deflate {
-            compression.clear();
-            for i in 3..10 {
-                compression.insert(i);
-            }
+        self.fast_evaluation = false;
+        self.filter.insert(RowFilter::Up);
+        self.filter.insert(RowFilter::MinSum);
+        self.filter.insert(RowFilter::BigEnt);
+        self.filter.insert(RowFilter::Brute);
+        if let Deflaters::Libdeflater { compression } = &mut self.deflate {
+            *compression = 12;
         }
-        self.apply_preset_4()
+        self
     }
 
     fn apply_preset_6(mut self) -> Self {
-        if let Deflaters::Zlib { compression, .. } = &mut self.deflate {
-            compression.clear();
-            for i in 1..10 {
-                compression.insert(i);
-            }
-        }
-        self.apply_preset_4()
+        self.filter.insert(RowFilter::Average);
+        self.filter.insert(RowFilter::Paeth);
+        self.apply_preset_5()
     }
 }
 
 impl Default for Options {
     fn default() -> Options {
         // Default settings based on -o 2 from the CLI interface
-        let mut filter = IndexSet::new();
-        filter.insert(0);
-        filter.insert(5);
-        let mut compression = IndexSet::new();
-        compression.insert(9);
-        let mut strategies = IndexSet::new();
-        for i in 0..4 {
-            strategies.insert(i);
-        }
-        // We always need NoOp to be present
-        let mut alphas = IndexSet::new();
-        alphas.insert(AlphaOptim::NoOp);
-
         Options {
             backup: false,
             check: false,
@@ -308,21 +293,17 @@ impl Default for Options {
             fix_errors: false,
             force: false,
             preserve_attrs: false,
-            filter,
+            filter: indexset! {RowFilter::None, RowFilter::Sub, RowFilter::Entropy, RowFilter::Bigrams},
             interlace: None,
-            alphas,
+            alphas: IndexSet::new(),
             bit_depth_reduction: true,
             color_type_reduction: true,
             palette_reduction: true,
             grayscale_reduction: true,
             idat_recoding: true,
             strip: Headers::None,
-            deflate: Deflaters::Zlib {
-                compression,
-                strategies,
-                window: 15,
-            },
-            use_heuristics: false,
+            deflate: Deflaters::Libdeflater { compression: 11 },
+            fast_evaluation: true,
             timeout: None,
         }
     }
@@ -470,10 +451,10 @@ pub fn optimize_from_memory(data: &[u8], opts: &Options) -> PngResult<Vec<u8>> {
 #[derive(Debug, PartialEq, PartialOrd, Clone, Copy)]
 /// Defines options to be used for a single compression trial
 struct TrialOptions {
-    pub filter: u8,
+    pub filter: RowFilter,
     pub compression: u8,
-    pub strategy: u8,
 }
+type TrialWithData = (TrialOptions, Vec<u8>);
 
 /// Perform optimization on the input PNG object using the options provided
 fn optimize_png(
@@ -482,10 +463,6 @@ fn optimize_png(
     opts: &Options,
     deadline: Arc<Deadline>,
 ) -> PngResult<Vec<u8>> {
-    type TrialWithData = (TrialOptions, Vec<u8>);
-
-    let original_png = png.clone();
-
     // Print png info
     let file_original_size = original_data.len();
     let idat_original_size = png.idat_data.len();
@@ -510,174 +487,132 @@ fn optimize_png(
     info!("    IDAT size = {} bytes", idat_original_size);
     info!("    File size = {} bytes", file_original_size);
 
-    let mut filter = opts.filter.clone();
-    let mut strategies = match &opts.deflate {
-        Deflaters::Zlib { strategies, .. } => Some(strategies.clone()),
-        _ => None,
-    };
-
-    if opts.use_heuristics {
-        // Heuristically determine which set of options to use
-        let (use_filter, use_strategy) = if png.raw.ihdr.bit_depth.as_u8() >= 8
-            && png.raw.ihdr.color_type != colors::ColorType::Indexed
-        {
-            (5, 1)
-        } else {
-            (0, 0)
-        };
-        if filter.is_empty() {
-            filter.insert(use_filter);
-        }
-        if let Some(strategies) = &mut strategies {
-            if strategies.is_empty() {
-                strategies.insert(use_strategy);
-            }
-        }
-    }
+    // Do this first so that reductions can ignore certain chunks such as bKGD
+    perform_strip(png, opts);
+    let stripped_png = png.clone();
 
+    // Must use normal (lazy) compression, as faster ones (greedy) are not representative
+    // Alpha reductions can benefit from higher compression but otherwise it's not beneficial
+    let eval_compression = 5;
+    // None and Bigrams work well together, especially for alpha reductions
+    let eval_filters = indexset! {RowFilter::None, RowFilter::Bigrams};
     // This will collect all versions of images and pick one that compresses best
-    let eval = Evaluator::new(deadline.clone());
-    // Usually we want transformations that are smaller than the unmodified original,
-    // but if we're interlacing, we have to accept a possible file size increase.
-    if opts.interlace.is_none() {
-        eval.set_baseline(png.raw.clone());
-    }
+    let eval = Evaluator::new(deadline.clone(), eval_filters.clone(), eval_compression);
     perform_reductions(png.raw.clone(), opts, &deadline, &eval);
-    let reduction_occurred = if let Some(result) = eval.get_result() {
-        *png = result;
-        true
+    let (reduction_occurred, mut eval_filter) = if let Some(result) = eval.get_best_candidate() {
+        *png = result.image;
+        (result.is_reduction, Some(result.filter))
     } else {
-        false
+        (false, None)
     };
 
     if opts.idat_recoding || reduction_occurred {
-        // Go through selected permutations and determine the best
-        let combinations = if let Deflaters::Zlib { compression, .. } = &opts.deflate {
-            filter.len() * compression.len() * strategies.as_ref().unwrap().len()
-        } else {
-            filter.len()
-        };
-        let mut results: Vec<TrialOptions> = Vec::with_capacity(combinations);
+        let mut filters = opts.filter.clone();
+        let fast_eval = opts.fast_evaluation && (filters.len() > 1 || eval_filter.is_some());
+        let best: Option<TrialWithData> = if fast_eval {
+            // Perform a fast evaluation of selected filters followed by a single main compression trial
 
-        for f in &filter {
-            if let Deflaters::Zlib { compression, .. } = &opts.deflate {
-                for zc in compression {
-                    for zs in strategies.as_ref().unwrap() {
-                        results.push(TrialOptions {
-                            filter: *f,
-                            compression: *zc,
-                            strategy: *zs,
-                        });
-                    }
-                    if deadline.passed() {
-                        break;
-                    }
+            if eval_filter.is_some() {
+                // Some filters have already been evaluated, we don't need to try them again
+                filters = filters.difference(&eval_filters).cloned().collect();
+            }
+
+            if !filters.is_empty() {
+                debug!("Evaluating: {} filters", filters.len());
+                let eval = Evaluator::new(deadline, filters, eval_compression);
+                if eval_filter.is_some() {
+                    eval.set_best_size(png.idat_data.len());
+                }
+                eval.try_image(png.raw.clone());
+                if let Some(result) = eval.get_best_candidate() {
+                    *png = result.image;
+                    eval_filter = Some(result.filter);
+                }
+            }
+
+            let trial = TrialOptions {
+                filter: eval_filter.unwrap(),
+                compression: match opts.deflate {
+                    Deflaters::Libdeflater { compression } => compression,
+                    _ => 0,
+                },
+            };
+            if trial.compression <= eval_compression {
+                // No further compression required
+                if png.idat_data.len() < idat_original_size || opts.force {
+                    Some((trial, png.idat_data.clone()))
+                } else {
+                    None
                 }
             } else {
-                // Zopfli and Libdeflater compression have no additional options.
+                info!("Trying: {}", trial.filter);
+                let original_len = idat_original_size;
+                let best_size = AtomicMin::new(if opts.force { None } else { Some(original_len) });
+                perform_trial(&png.filtered, opts, trial, &best_size)
+            }
+        } else {
+            // Perform full compression trials of selected filters and determine the best
+
+            if filters.is_empty() {
+                // Pick a filter automatically
+                if png.raw.ihdr.bit_depth.as_u8() >= 8 {
+                    // Bigrams is the best all-rounder when there's at least one byte per pixel
+                    filters.insert(RowFilter::Bigrams);
+                } else {
+                    // Otherwise delta filters generally don't work well, so just stick with None
+                    filters.insert(RowFilter::None);
+                }
+            }
+
+            let mut results: Vec<TrialOptions> = Vec::with_capacity(filters.len());
+
+            for f in &filters {
                 results.push(TrialOptions {
                     filter: *f,
-                    compression: 0,
-                    strategy: 0,
+                    compression: match opts.deflate {
+                        Deflaters::Libdeflater { compression } => compression,
+                        _ => 0,
+                    },
                 });
             }
 
-            if deadline.passed() {
-                break;
-            }
-        }
+            info!("Trying: {} filters", results.len());
 
-        info!("Trying: {} combinations", results.len());
-
-        let filters: IndexMap<u8, Vec<u8>> = filter
-            .par_iter()
-            .with_max_len(1)
-            .map(|f| {
-                let png = png.clone();
-                (*f, png.raw.filter_image(*f))
-            })
-            .collect();
-
-        let original_len = original_png.idat_data.len();
-        let added_interlacing = opts.interlace == Some(1) && original_png.raw.ihdr.interlaced == 0;
-
-        let best_size = AtomicMin::new(if opts.force { None } else { Some(original_len) });
-        let results_iter = results.into_par_iter().with_max_len(1);
-        let best = results_iter.filter_map(|trial| {
-            if deadline.passed() {
-                return None;
-            }
-            let filtered = &filters[&trial.filter];
-            let new_idat = match opts.deflate {
-                Deflaters::Zlib { window, .. } => deflate::deflate(
-                    filtered,
-                    trial.compression,
-                    trial.strategy,
-                    window,
-                    &best_size,
-                    &deadline,
-                ),
-                #[cfg(feature = "zopfli")]
-                Deflaters::Zopfli { iterations } => deflate::zopfli_deflate(filtered, iterations),
-                #[cfg(feature = "libdeflater")]
-                Deflaters::Libdeflater => deflate::libdeflater_deflate(filtered, &best_size),
-            };
-
-            let new_idat = match new_idat {
-                Ok(n) => n,
-                Err(PngError::DeflatedDataTooLong(max)) => {
-                    debug!(
-                        "    zc = {}  zs = {}  f = {}       >{} bytes",
-                        trial.compression, trial.strategy, trial.filter, max,
-                    );
+            let original_len = idat_original_size;
+            let best_size = AtomicMin::new(if opts.force { None } else { Some(original_len) });
+            let results_iter = results.into_par_iter().with_max_len(1);
+            let best = results_iter.filter_map(|trial| {
+                if deadline.passed() {
                     return None;
                 }
-                Err(_) => return None,
-            };
-
-            // update best size across all threads
-            let new_size = new_idat.len();
-            best_size.set_min(new_size);
-
-            debug!(
-                "    zc = {}  zs = {}  f = {}        {} bytes",
-                trial.compression,
-                trial.strategy,
-                trial.filter,
-                new_idat.len()
-            );
-
-            if new_size < original_len || added_interlacing || opts.force {
-                Some((trial, new_idat))
-            } else {
-                None
-            }
-        });
-        let best: Option<TrialWithData> = best.reduce_with(|i, j| {
-            if i.1.len() < j.1.len() || (i.1.len() == j.1.len() && i.0 < j.0) {
-                i
-            } else {
-                j
-            }
-        });
+                let filtered = &png.raw.filter_image(trial.filter);
+                perform_trial(filtered, opts, trial, &best_size)
+            });
+            best.reduce_with(|i, j| {
+                if i.1.len() < j.1.len() || (i.1.len() == j.1.len() && i.0 < j.0) {
+                    i
+                } else {
+                    j
+                }
+            })
+        };
 
         if let Some((opts, idat_data)) = best {
             png.idat_data = idat_data;
             info!("Found better combination:");
             info!(
-                "    zc = {}  zs = {}  f = {}        {} bytes",
+                "    zc = {}  f = {}  {} bytes",
                 opts.compression,
-                opts.strategy,
                 opts.filter,
                 png.idat_data.len()
             );
-        } else if reduction_occurred {
-            *png = original_png;
+        } else if eval_filter.is_some() {
+            *png = stripped_png;
         }
+    } else if png.idat_data.len() >= idat_original_size {
+        *png = stripped_png;
     }
 
-    perform_strip(png, opts);
-
     let output = png.output();
 
     if idat_original_size >= png.idat_data.len() {
@@ -739,11 +674,17 @@ fn perform_reductions(
     deadline: &Deadline,
     eval: &Evaluator,
 ) {
+    // The eval baseline will be set from the original png only if we attempt any reductions
+    let mut baseline = Some(png.clone());
+    let mut reduction_occurred = false;
+
     // must be done first to evaluate rest with the correct interlacing
     if let Some(interlacing) = opts.interlace {
         if let Some(reduced) = png.change_interlacing(interlacing) {
             png = Arc::new(reduced);
             eval.try_image(png.clone());
+            // If we're interlacing, we have to accept a possible file size increase
+            baseline = None;
         }
         if deadline.passed() {
             return;
@@ -755,6 +696,7 @@ fn perform_reductions(
             png = Arc::new(reduced);
             eval.try_image(png.clone());
             report_reduction(&png);
+            reduction_occurred = true;
         }
         if deadline.passed() {
             return;
@@ -776,6 +718,7 @@ fn perform_reductions(
                 }
             }
             report_reduction(&png);
+            reduction_occurred = true;
         }
         if deadline.passed() {
             return;
@@ -783,17 +726,72 @@ fn perform_reductions(
     }
 
     if opts.color_type_reduction {
+        // Perform a black alpha reduction before color type reductions
+        // This can allow reductions from alpha to indexed which may not have been possible otherwise
+        if !opts.alphas.is_empty() {
+            if let Some(reduced) = filtered_alpha_channel(&png, AlphaOptim::Black) {
+                png = Arc::new(reduced);
+            }
+        }
         if let Some(reduced) = reduce_color_type(&png, opts.grayscale_reduction) {
             png = Arc::new(reduced);
             eval.try_image(png.clone());
             report_reduction(&png);
+            reduction_occurred = true;
         }
         if deadline.passed() {
             return;
         }
     }
 
-    try_alpha_reductions(png, &opts.alphas, eval);
+    if try_alpha_reductions(png, &opts.alphas, eval) {
+        reduction_occurred = true;
+    }
+
+    if let Some(baseline) = baseline {
+        if reduction_occurred {
+            eval.set_baseline(baseline);
+        }
+    }
+}
+
+/// Execute a compression trial
+fn perform_trial(
+    filtered: &[u8],
+    opts: &Options,
+    trial: TrialOptions,
+    best_size: &AtomicMin,
+) -> Option<TrialWithData> {
+    let new_idat = match opts.deflate {
+        Deflaters::Libdeflater { .. } => deflate::deflate(filtered, trial.compression, best_size),
+        #[cfg(feature = "zopfli")]
+        Deflaters::Zopfli { iterations } => deflate::zopfli_deflate(filtered, iterations),
+    };
+
+    // update best size or convert to error if not smaller
+    let new_idat = match new_idat {
+        Ok(n) if !best_size.set_min(n.len()) => Err(PngError::DeflatedDataTooLong(n.len())),
+        _ => new_idat,
+    };
+
+    match new_idat {
+        Ok(n) => {
+            let bytes = n.len();
+            debug!(
+                "    zc = {}  f = {}  {} bytes",
+                trial.compression, trial.filter, bytes
+            );
+            Some((trial, n))
+        }
+        Err(PngError::DeflatedDataTooLong(bytes)) => {
+            debug!(
+                "    zc = {}  f = {} >{} bytes",
+                trial.compression, trial.filter, bytes,
+            );
+            None
+        }
+        Err(_) => None,
+    }
 }
 
 #[derive(Debug)]
@@ -878,10 +876,8 @@ fn perform_strip(png: &mut PngData, opts: &Options) {
             }
         }
         Headers::Safe => {
-            const PRESERVED_HEADERS: [[u8; 4]; 9] = [
-                *b"cHRM", *b"gAMA", *b"iCCP", *b"sBIT", *b"sRGB", *b"bKGD", *b"hIST", *b"pHYs",
-                *b"sPLT",
-            ];
+            const PRESERVED_HEADERS: [[u8; 4]; 5] =
+                [*b"cICP", *b"iCCP", *b"sBIT", *b"sRGB", *b"pHYs"];
             let keys: Vec<[u8; 4]> = raw.aux_headers.keys().cloned().collect();
             for hdr in &keys {
                 if !PRESERVED_HEADERS.contains(hdr) {
@@ -933,7 +929,9 @@ fn srgb_rendering_intent(mut iccp: &[u8]) -> Option<u8> {
     if compression_method != 0 {
         return None; // The profile is supposed to be compressed (method 0)
     }
-    let icc_data = inflate(compressed_data).ok()?;
+    // The decompressed size is unknown so we have to guess the required buffer size
+    let max_size = (compressed_data.len() * 2).max(1000);
+    let icc_data = inflate(compressed_data, max_size).ok()?;
 
     let rendering_intent = *icc_data.get(67)?;
 
@@ -949,10 +947,7 @@ fn srgb_rendering_intent(mut iccp: &[u8]) -> Option<u8> {
         }
         b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00" => {
             // Known-bad profiles are identified by their CRC
-            match (
-                Crc::<u32>::new(&CRC_32_ISO_HDLC).checksum(&icc_data),
-                icc_data.len(),
-            ) {
+            match (crc32(&icc_data), icc_data.len()) {
                 (0x5d51_29ce, 3024) | (0x182e_a552, 3144) | (0xf29e_526d, 3144) => {
                     Some(rendering_intent)
                 }
